from .with_parameters import WithParameters
from .optimizable_parameter import OptimizableParameter
from .fixed_parameter import FixedParameter

from .log_parameter_transformation import LogParameterTransformation
from .none_parameter_transformation import NoneParameterTransformation


# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


class DataTransformation(object):
    """

    """

    def __init__(self, data):

        self.transform_matrices_dict = {
            key: DataTransformation._get_transform_matrices(value)
            for key, value in data.items()
        }

    @staticmethod
    def _get_transform_matrices(matrix):
        """

        :param matrix:
        :type matrix:
        :return:
        :rtype:
        """
        min_x = np.min(matrix, axis=0)
        scale_inv = np.max(matrix, axis=0) - min_x
        scale = np.power(scale_inv, -1)
        scale_inv = np.diag(scale_inv)
        scale = np.diag(scale)
        translate_inv = min_x
        translate = -np.dot(min_x, scale)
        return np.vstack((scale, translate)), \
            np.vstack((scale_inv, translate_inv))

    def transform(self, data, inv=False):
        """

        :param data:
        :type data:
        :param transform:
        :type transform:
        :param inv:
        :type inv:
        :return:
        :rtype:
        """

        return {
            key: self.transform_key(value, key, inv)
            for key, value in data.items()
        }

    def scale(self, data, inv=False):
        """

        :param data:
        :type data:
        :param transform:
        :type transform:
        :param inv:
        :type inv:
        :return:
        :rtype:
        """

        return {
            key: self.scale_key(value, key, inv)
            for key, value in data.items()
        }

    def transform_key(self, matrix, key, inv=False):
        """

        :param matrix:
        :type matrix:
        :param key:
        :type key:
        :param inv:
        :type inv:
        :return:
        :rtype:
        """

        return np.dot(
            np.hstack((matrix, np.ones((len(matrix), 1)))),
            self.transform_matrices_dict[key][1 if inv else 0]
        )

    def scale_key(self, matrix, key, inv=False):
        """

        :param matrix:
        :type matrix:
        :param key:
        :type key:
        :param inv:
        :type inv:
        :return:
        :rtype:
        """

        return np.dot(
            matrix,
            self.transform_matrices_dict[key][1 if inv else 0][:-1]
        )



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .stationary_function import StationaryFunction
from ..parameters.optimizable_parameter import OptimizableParameter
from ..parameters.log_parameter_transformation import LogParameterTransformation


class Exponential(StationaryFunction):
    """

    """
    def __init__(self, ov2=1, ls=None, ard_dims=None):
        hyperparams = [
            OptimizableParameter(
                'output_variance', LogParameterTransformation,
                default_value=ov2,
                min_value=np.exp(-20), max_value=np.exp(20)
            )
        ]

        super(Exponential, self).__init__(hyperparams, ls, ard_dims)

    def stationary_function(self, sq_dist):
        """
        It applies the exponential kernel function
        element-wise to the distance matrix.

        .. math::
            k_{E}(r) = \exp (-\dfrac{r}{l})

        :param sq_dist: Distance matrix
        :type sq_dist:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """
        dist = np.sqrt(sq_dist)
        return np.exp(-dist) * \
            np.square(self.get_param_value('output_variance'))

    def dkr_dx(self, sq_dist, dr_dx):
        """
        Measures gradient of the kernel function in X.

        :param sq_dist: Square distance
        :type sq_dist:
        :param dr_dx:
        :type dr_dx:
        :return: 3D array with the gradient of the kernel function in every
         dimension of X.
        :rtype:
        """
        sq_power = sq_dist.copy()
        sq_power[sq_power != 0.0] = np.power(sq_power[sq_power != 0.0], -0.5)
        result = -0.5 * sq_power[:, :, np.newaxis] * \
                 np.exp(-np.power(sq_dist, 0.5))[:, :, np.newaxis] * dr_dx
        return result * np.square(self.get_param_value('output_variance'))

    def dkr_dtheta(self, sq_dist, trans):
        """
        Measures gradient of the kernel function in the
        hyper-parameter space.

        :param sq_dist: Square distance
        :type sq_dist:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient of the kernel function in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """
        dist = np.sqrt(sq_dist)
        sq_power = sq_dist.copy()
        sq_power[sq_power != 0.0] = np.power(sq_power[sq_power != 0.0], -0.5)

        dkr_dov = np.exp(-dist) * \
            2.0 * self.get_param_value('output_variance')

        if trans:
            dkr_dov = self.get_hyperparam('output_variance').grad_trans(
                dkr_dov
            )

        return dkr_dov,

    def dkr_dl(self, sq_dist, dr_dl):
        """
        Measures gradient of the kernel function in the
        hyper-parameter space.

        :param sq_dist: Square distance
        :type sq_dist:
        :param dr_dl:
        :type dr_dl:
        :return: 3D array with the gradient of the kernel function in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """
        sq_power = sq_dist.copy()
        sq_power[sq_power != 0.0] = np.power(sq_power[sq_power != 0.0], -0.5)
        result = -0.5 * sq_power[:, :, np.newaxis] * \
                 np.exp(-np.power(sq_dist, 0.5))[:, :, np.newaxis] * dr_dl
        return result * np.square(self.get_param_value('output_variance'))



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .covariance_function import CovarianceFunction
from ..parameters.optimizable_parameter import OptimizableParameter
from ..parameters.log_parameter_transformation import LogParameterTransformation


class WhiteNoise(CovarianceFunction):
    """

    """
    def __init__(self, ov2=1e-4):

        hyperparams = [
            OptimizableParameter(
                'output_variance', LogParameterTransformation,
                default_value=ov2,
                min_value=np.exp(-15), max_value=np.exp(5)
            )
        ]

        super(WhiteNoise, self).__init__(hyperparams)

    def covariance(self, mat_a, mat_b=None, only_diagonal=False):
        """
        Measures the distance matrix between solutions of A and B, and
        applies the kernel function element-wise to the distance matrix.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param only_diagonal:
        :type only_diagonal:
        :return: Result matrix with kernel function applied element-wise.
        :rtype:
        """
        len_a = len(mat_a)
        if mat_b is not None:
            len_b = len(mat_b)

            return np.zeros((len_a, len_b))

        if only_diagonal:
            return np.square(self.get_param_value('output_variance')) * \
                np.ones((len_a, 1))

        return np.square(self.get_param_value('output_variance')) * \
            np.eye(len_a)

    def dk_dx(self, mat_a, mat_b=None):
        """
        Measures gradient of the distance between solutions of A and B in X.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :param mat_b: List of solutions in lines and dimensions in columns.
        :return: 3D array with the gradient in every dimension of X.
        """

        raise NotImplementedError("Not Implemented. This is an interface.")

    def dk_dtheta(self, mat_a, mat_b=None, trans=False):
        """
        Measures gradient of the distance between solutions of A and B in the
        hyper-parameter space.

        :param mat_a: List of solutions in lines and dimensions in columns.
        :type mat_a:
        :param mat_b: List of solutions in lines and dimensions in columns.
        :type mat_b:
        :param trans: Return results in the transformed space.
        :type trans:
        :return: 3D array with the gradient in every
         dimension the length-scale hyper-parameter space.
        :rtype:
        """

        len_a = len(mat_a)
        if mat_b is not None:
            len_b = len(mat_b)

            return np.zeros((len_a, len_b)),

        dk_dov = np.eye(len_a) * \
            2.0 * self.get_param_value('output_variance')

        if trans:
            dk_dov = self.get_hyperparam('output_variance').grad_trans(dk_dov)

        return dk_dov,



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .parametrizable import Parametrizable


class Parameter(Parametrizable):
    """

    """
    def __init__(self, name, transformation, default_value):
        """

        :param name:
        :type name:
        :param transformation:
        :type transformation:
        :param default_value:
        :type default_value:
        """

        self.name = name
        self.transformation = transformation
        self.default_value = default_value
        self.array = hasattr(self.default_value, "__len__")
        self.current_value = self.default_value

        self.dims = 1
        if self.array:
            self.dims = len(self.default_value)

    def is_array(self):
        """

        :return:
        :rtype:
        """
        return self.array

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        assert len(params) == self.dims, \
            "length of {} is not correct".format(self.name)

        if trans:
            if self.array:
                params = self.transformation.inv_trans(params).tolist()
            else:
                params = self.transformation.inv_trans(params)

        if self.array is False:
            self.current_value = params[0]
        else:
            self.current_value = params

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        self.current_value = self.default_value

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        assert self.current_value is not None, \
            "{} has not been initialized".format(self.name)

        current_value = self.current_value
        if trans:
            current_value = self.transformation.trans(current_value)
            if self.array:
                current_value = current_value.tolist()

        if self.array:
            return current_value
        return [current_value]

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if not recursive:
            return self.name

        if self.dims == 1:
            return [self.name]

        return [
            "{}_d{}".format(self.name, dim) for dim in range(self.dims)
        ]

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        raise NotImplementedError("Not Implemented. This is an interface.")

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        return self.dims



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .parameter import Parameter


class OptimizableParameter(Parameter):
    """

    """
    def __init__(self, name, transformation, default_value=1.0,
                 min_value=-np.inf, max_value=np.inf,
                 jitter_sd=0.1):
        """

        :param name:
        :type name:
        :param transformation:
        :type transformation:
        :param default_value:
        :type default_value:
        :param min_value:
        :type min_value:
        :param max_value:
        :type max_value:
        :param jitter_sd:
        :type jitter_sd:
        """

        super(OptimizableParameter, self).__init__(
            name, transformation, default_value
        )

        self.min_value = np.float64(min_value)
        self.max_value = np.float64(max_value)

        assert (np.all(self.min_value <= default_value) and
                np.all(default_value <= self.max_value)),\
            "{} is out of bounds".format(self.name)

        self.optimized_value = None

        self.jitter_sd = jitter_sd

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        super(OptimizableParameter, self).set_param_values(
            params, optimizable_only=optimizable_only, trans=trans
        )

        assert (np.all(self.min_value <= self.current_value) and
                np.all(self.current_value <= self.max_value)), \
            "{} is out of bounds".format(self.name)

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """

        self.optimized_value = None

        super(OptimizableParameter, self).set_params_to_default(
            optimizable_only=optimizable_only
        )

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        min_value = self.min_value
        max_value = self.max_value
        if trans:
            min_value = self.transformation.trans(min_value)
            max_value = self.transformation.trans(max_value)

        if self.optimized_value is not None:
            optimized_value = self.optimized_value
            if trans:
                optimized_value = self.transformation.trans(optimized_value)
            current_value = None
            while not (current_value is not None and
                       np.all(min_value < current_value) and
                       np.all(current_value < max_value)):
                current_value = optimized_value + \
                    np.array(
                        np.random.normal(
                            loc=0.0,
                            scale=self.jitter_sd,
                            size=self.dims
                        )
                    )
        else:
            current_value = np.array(np.random.uniform(
                min_value, max_value, self.dims
            ))

        self.set_param_values(current_value, trans=trans)

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        self.optimized_value = self.current_value

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        min_value = self.min_value
        max_value = self.max_value

        if trans:
            min_value = self.transformation.trans(min_value)
            max_value = self.transformation.trans(max_value)

        return [(min_value, max_value)] * self.dims

    def set_min_value(self, min_value):
        """

        :param min_value:
        :type min_value:
        :return:
        :rtype:
        """
        self.min_value = min_value

    def set_max_value(self, max_value):
        """

        :param max_value:
        :type max_value:
        :return:
        :rtype:
        """
        self.max_value = max_value

    def grad_trans(self, df):
        """

        :param df:
        :type df:
        :return:
        :rtype:
        """

        return self.transformation.grad_trans(self.current_value, df)



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

from .parameter import Parameter


class FixedParameter(Parameter):
    """

    """

    def set_param_values(self, params, optimizable_only=False, trans=False):
        """

        :param params:
        :type params:
        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if optimizable_only:
            return

        super(FixedParameter, self).set_param_values(
            params, optimizable_only=optimizable_only, trans=trans
        )

    def set_params_to_default(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return

        super(FixedParameter, self).set_params_to_default(
            optimizable_only=optimizable_only
        )

    def set_params_at_random(self, trans=False):
        """

        :return:
        :rtype:
        """
        pass

    def save_current_as_optimized(self):
        """

        :return:
        :rtype:
        """
        pass

    def get_param_values(self, optimizable_only=False, trans=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :param trans:
        :type trans:
        :return:
        :rtype:
        """
        if optimizable_only:
            return []

        return super(FixedParameter, self).get_param_values(
            optimizable_only=optimizable_only, trans=trans
        )

    def get_param_keys(self, recursive=True, optimizable_only=False):
        """

        :param recursive:
        :type recursive:
        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return []

        return super(FixedParameter, self).get_param_keys(
            recursive=recursive,
            optimizable_only=optimizable_only
        )

    def get_param_bounds(self, trans=False):
        """

        :param trans:
        :type trans:
        :return:
        :rtype:
        """

        return []

    def get_param_n(self, optimizable_only=False):
        """

        :param optimizable_only:
        :type optimizable_only:
        :return:
        :rtype:
        """
        if optimizable_only:
            return 0

        return super(FixedParameter, self).get_param_n(
            optimizable_only=optimizable_only
        )



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import copy

import numpy as np

from .parameters.with_parameters import WithParameters
from .covariance_functions.posterior import Posterior as PosteriorCov
from .mean_functions.posterior import Posterior as PosteriorMean


class GaussianProcess(WithParameters):
    """
    GPlib module for Gaussian Process
    """
    def __init__(self,
                 mean_function,
                 covariance_function,
                 likelihood_function,
                 inference_method):
        self.mean_function = mean_function
        self.covariance_function = covariance_function
        self.likelihood_function = likelihood_function
        self.inference_method = inference_method

        self.likelihood_function.set_gp(self)
        self.inference_method.set_gp(self)

        super(GaussianProcess, self).__init__([
            mean_function, covariance_function
        ])

    def get_posterior(self, data):
        """

        :param data:
        :type data:
        :return:
        :rtype:
        """

        # Create a copy of self to be the prior of the posterior GP
        prior_gp = copy.deepcopy(self)
        data_copy = copy.deepcopy(data)

        return GaussianProcess(
            PosteriorMean(prior_gp, data_copy),
            PosteriorCov(prior_gp, data_copy),
            # Create a copy of the prior to be the likelihood of the posterior
            copy.copy(prior_gp.likelihood_function),
            # Create a copy of the prior to be the inference of the posterior
            copy.copy(prior_gp.inference_method)
        )

    def sample(self, test_points, n_samples=10):
        """
        Sample the prior of the GP

        :param test_points:
        :type test_points:
        :param n_samples:
        :type n_samples:
        :return:
        :rtype:
        """
        marginal = self.inference_method.marginalize_gp(test_points)
        rnd = np.random.randn(test_points.shape[0], n_samples)

        return marginal['mean'].reshape(-1, 1) + \
            np.dot(marginal['l_matrix'], rnd)



# -*- coding: utf-8 -*-
#
#    Copyright 2019 Ibai Roman
#
#    This file is part of GPlib.
#
#    GPlib is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GPlib is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GPlib. If not, see <http://www.gnu.org/licenses/>.

import os.path
import setuptools


def read(file):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, file), 'r') as f:
        return f.read()


setuptools.setup(
    name='gplib',
    version='0.11.5',
    author='Ibai Roman',
    author_email='ibaidev@protonmail.com',
    description='Python library for Gaussian Process Regression.',
    license='GPLv3',
    keywords='Gaussian Process',
    url='https://gitlab.com/ibaidev/gplib',
    long_description=read('README.rst'),
    packages=setuptools.find_packages(exclude=['contrib', 'docs', 'tests']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    install_requires=[
        'scipy',
        'numpy',
        'matplotlib'
    ]
)
